# -*- coding: utf-8 -*-
"""
Helpers for nice shell output formatting.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import re


def cstr(msg):
    """ Format color marked string into proper shell opcodes. """
    return re.sub(r'\^(\d{1,2})', '\x1b[\\1m', msg)


def cprint(msg, *args, **kw):
    """ Print colored message to stdout. """
    if len(args) or len(kw):
        msg = msg.format(*args, **kw)

    print(cstr(msg))


def info(msg, *args, **kw):
    """ Print sys message to stdout.

    System messages should inform about the flow of the script. This should
    be a major milestones during the build.
    """
    if len(args) or len(kw):
        msg = msg.format(*args, **kw)

    print(cstr('-- ^32{}^0'.format(msg)))


def err(msg, *args, **kw):
    """ Per step status messages

    Use this locally in a command definition to highlight more important
    information.
    """
    if len(args) or len(kw):
        msg = msg.format(*args, **kw)

    print(cstr('-- ^31{}^0'.format(msg)))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

# stdlib imports
import os
from os.path import expanduser, exists, join

# 3rd party imports
from cached_property import cached_property_ttl

# local imports
from peltak.core import log
from .scaffold import Scaffold


class LocalStore(object):
    DEFAULT_PATH = expanduser('~/.config/peltak/local')

    def __init__(self, path=None):
        self.path = path or LocalStore.DEFAULT_PATH
        if not exists(self.path):
            os.makedirs(self.path)

    @cached_property_ttl(ttl=5)
    def scaffolds(self):
        scaffolds = []
        for filename in os.listdir(self.path):
            if filename.endswith(Scaffold.FILE_EXT):
                try:
                    path = join(self.path, filename)
                    scaffold = Scaffold.load_from_file(path)

                    scaffolds.append(scaffold)
                except Scaffold.Invalid:
                    pass

        return scaffolds

    def add(self, scaffold):
        scaffold.write(self.path)

        self._invalidate_cache()

    def delete(self, name):
        scaffold = self.load(name)

        if scaffold is not None:
            os.remove(scaffold.path)
            self._invalidate_cache()

        else:
            log.err("'{}' does not exist".format(name))

    def load(self, name):
        return next((x for x in self.scaffolds if x.name == name), None)

    def push(self, name):
        raise NotImplemented("Remote service is not yet implemented")

    def pull(self, name):
        raise NotImplemented("Remote service is not yet implemented")

    def _invalidate_cache(self):
        if 'scaffolds' in self.__dict__:
            del self.__dict__['scaffolds']



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

# local imports
from peltak.core import conf
from peltak.commands import cli
__all__ = [
    'cli'
]


conf.load()


try:
    # Only enabled click_completion if psutil package is installed
    import psutil
    import click_completion
    click_completion.init()
except ImportError:
    pass


from peltak.commands import scaffold
from peltak.commands import clean
from peltak.commands import git



# -*- coding: utf-8 -*-
"""
Django related management commands.

Those commands replace the usage of ``./manage.py`` (thus it's removed). Those
correspond 1 to 1 to their ``./manage.py`` counterparts but the arguments are
in the fabric format (can't get around this).
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import sys
import click
from os import environ

# local imports
from peltak.core import conf
from . import cli


@cli.group()
def django():
    """ Commands related to django """
    pass


def _manage_cmd(cmd, settings=None):
    """ Run django ./manage.py command manually.

    This function eliminates the need for having ``manage.py`` (reduces file
    clutter).
    """
    sys.path.insert(0, conf.get('SRC_DIR'))

    settings = settings or conf.get('DJANGO_SETTINGS', None)
    environ.setdefault("DJANGO_SETTINGS_MODULE", settings)

    from django.core.management import execute_from_command_line

    args = sys.argv[0:-1] + cmd

    execute_from_command_line(args)


@django.command()
@click.option('-p', '--port', type=int, default=8000)
@click.option('-s', '--settings', type=str)
def devserver(port=8000, settings=None):
    """ Run dev server. """
    _manage_cmd(['runserver', '0.0.0.0:{}'.format(port)], settings)


@django.command()
def collectstatic():
    """ Collect all static files. """
    _manage_cmd(['collectstatic', '--no-input'])


@django.command()
@click.argument('app')
@click.argument('name')
def mkmigrations(app, name):
    """ Create migrations for a given app. """
    _manage_cmd(['makemigrations', '-n', name] + app.split(' '))


@django.command()
def migrate():
    """ Apply pending migrations. """
    _manage_cmd(['migrate'])


@django.command()
def createsuperuser():
    """ Create super user (probably needed for the first user). """
    _manage_cmd(['createsuperuser'])


@django.command()
def shell():
    """ Start django shell """
    _manage_cmd(['shell'])



# -*- coding: utf-8 -*-
"""
Per-project configuration support
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import os
import subprocess
import sys
from collections import namedtuple
from contextlib import contextmanager
from os.path import abspath, dirname, isabs, join, normpath

# local imports
from . import log


g_config = {}
g_proj_path = None
g_proj_root = None
PROJ_CONF_FILE = 'pelconf.py'


ExecResult = namedtuple(
    'ExecResult',
    'command return_code stdout stderr succeeded failed'
)


def init(config):
    """ Initialize configuration with the given values.

    This should be called from within the project fabfile, before any
    other commands are imported

    :param dict config:
        The dictionary containing the project configuration.
    """
    global g_config

    g_config.update(config)


def load():
    if proj_path() is None:
        return

    with within_proj_dir():
        if sys.version_info >= (3, 5):
            from importlib.util import spec_from_file_location
            from importlib.util import module_from_spec

            spec = spec_from_file_location('pelconf', 'pelconf.py')
            mod = module_from_spec(spec)
            spec.loader.exec_module(mod)

        elif sys.version_info >= (3, 3):
            from importlib.machinery import SourceFileLoader
            loader = SourceFileLoader('pelconf', 'pelconf.py')
            mod = loader.load_module()

        elif sys.version_info <= (3, 0):
            import imp

            imp.load_source('pelconf', 'pelconf.py')


def getenv(name, default=None):
    """ Get the value of an ENV variable. """
    return os.environ.get(name, default)


def proj_path(path=None):
    """ Return absolute path to the repo dir (root project directory). """
    global g_proj_path

    path = path or '.'

    if not isabs(path):
        if g_proj_path is None:
            g_proj_path = _find_proj_root()

        if g_proj_path is not None:
            path = normpath(join(g_proj_path, path))

    return path


@contextmanager
def within_proj_dir(path='.', quiet=False):
    """ Return absolute path to the repo dir (root project directory). """
    curr_dir = os.getcwd()

    os.chdir(proj_path(path))

    yield

    os.chdir(curr_dir)


def run(cmd, capture=False, shell=True, env=None):
    options = {
        'bufsize': 1,       # line buffered
        'shell': shell
    }

    if capture:
        options.update({
            'stdout': subprocess.PIPE,
            'stderr': subprocess.PIPE,
        })

    if env is not None:
        options['env'] = dict(os.environ)
        options['env'].update(env)

    p = subprocess.Popen(cmd, **options)
    stdout, stderr = p.communicate()

    if capture is False and p.returncode != 0:
        sys.exit(p.returncode)

    return ExecResult(
        cmd,
        p.returncode,
        stdout and stdout.decode('utf-8'),
        stderr and stderr.decode('utf-8'),
        p.returncode == 0,
        p.returncode != 0
    )


def get(name, *default):
    """ Get config value with the given name and optional default.

    :param str|unicode name:
        The name of the config value.
    :param Any default:
        If given and the key doesn't not exist, this will be returned instead.
        If it's not given and the config value does not exist, AttributeError
        will be raised
    :return Any:
        The requested config value. This is one of the global values defined
        in this file.
    """
    global g_config

    if name in g_config:
        return g_config[name]
    elif default:
        return default[0]
    else:
        raise AttributeError("Config value '{}' does not exist".format(name))


def get_path(name, *default):
    """ Get config value as path relative to the project directory.

    This allows easily defining the project configuration within the fabfile
    as always relative to that fabfile.

    :param str|unicode name:
        The name of the config value containing the path.
    :param Any default:
        If given and the key doesn't not exist, this will be returned instead.
        If it's not given and the config value does not exist, AttributeError
        will be raised
    :return Any:
        The requested config value. This is one of the global values defined
        in this file.
    """
    global g_config

    if name in g_config:
        return proj_path(g_config[name])
    elif default:
        if default[0] is None:
            return None
        else:
            return proj_path(default[0])
    else:
        raise AttributeError("Config value '{}' does not exist".format(name))


def _find_proj_root():
    """ Find appengine_sdk in the current $PATH. """
    global g_proj_root

    if g_proj_root is None:
        start_paths = [
            abspath(dirname(__file__)),
            os.getcwd()
        ]

        # log.info('Finding project root')
        for curr in start_paths:
            while curr.startswith('/') and len(curr) > 1:
                # log.info('  checking ^94{}', curr)
                if PROJ_CONF_FILE in os.listdir(curr):
                    # log.info('  ^32Found')
                    g_proj_root = curr
                    break
                else:
                    curr = normpath(join(curr, '..'))

            if g_proj_root is not None:
                break

        if g_proj_root is None:
            # log.info('  ^31Not found')
            pass

    return g_proj_root



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

# package interface
from .local import LocalStore
from .scaffold import Scaffold
__all__ = [
    'LocalStore',
    'Scaffold'
]



# -*- coding: utf-8 -*-
"""
Git helpers.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
from collections import namedtuple
from os.path import join, exists

# 3rd party imports

# local imports
from . import conf


Author = namedtuple('Author', 'name email')


def current_branch():
    """ Return the name of the currently checked out git branch. """
    return conf.run('git symbolic-ref --short HEAD', capture=True).stdout


def is_dirty(path='.'):
    """ Return **True** if there are any changes/unstaged files. """

    with conf.within_proj_dir(path, quiet=True):
        status = conf.run('git status --porcelain', capture=True).stdout
        return bool(status.strip())


def commit_author(sha1=''):
    """ Return the author of the given commit.

    :param str sha1:
        The sha1 of the commit to query. If not given, it will return the sha1
        for the current commit.
    :return Author:
        A named tuple ``(name, email)`` with the commit author details.
    """
    with conf.within_proj_dir():
        cmd = 'git show -s --format="%an||%ae" {}'.format(sha1)
        result = conf.run(cmd, capture=True).stdout
        name, email = result.split('||')
        return Author(name, email)


def load_gitignore(repo_path):
    gitignore_path = join(repo_path, '.gitignore')

    exclude_patterns = []

    if exists(gitignore_path):
        with open(gitignore_path) as fp:
            for line in fp.readlines():
                line = line.strip()

                if line.startswith('#'):
                    continue

                elif line.startswith('/'):
                    line = line[1:]

                exclude_patterns.append(line)

    return exclude_patterns





# -*- coding: utf-8 -*-
"""
git helpers.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import os

# 3rd party imports
import click

# local imports
from peltak.core import conf
from peltak.core import git
from peltak.core import log
from . import cli


@cli.group('git')
def git_group():
    """ Git related commands """
    pass


@git_group.command('add-hooks')
def add_hooks():
    """ Setup project git hooks.

    This will run all the checks before pushing to avoid waiting for the CI
    fail.
    """
    log.info("Adding pre-commit hook")
    with open(conf.proj_path('.git/hooks/pre-commit'), 'w') as fp:
        fp.write('\n'.join([
            '#!/bin/bash',
            'PATH="/opt/local/libexec/gnubin:$PATH"',
            (
                'REPO_PATH=$(dirname "$(dirname "$(dirname '
                '"$(readlink -fm "$0")")")")'
            ),
            '',
            'echo "REPO_PATH=$REPO_PATH"',
            'source "$REPO_PATH/env/bin/activate"',
            '',
            'fab lint',
        ]))

    log.info("Adding pre-push hook")
    with open(conf.proj_path('.git/hooks/pre-push'), 'w') as fp:
        fp.write('\n'.join([
            '#!/bin/bash',
            'PATH="/opt/local/libexec/gnubin:$PATH"',
            (
                'REPO_PATH=$(dirname "$(dirname "$(dirname '
                '"$(readlink -fm "$0")")")")'
            ),
            '',
            'source "$REPO_PATH/env/bin/activate"',
            '',
            'fab test',
        ]))

    log.info("Making hooks executable")
    os.chmod(conf.proj_path('.git/hooks/pre-commit'), 0o755)
    os.chmod(conf.proj_path('.git/hooks/pre-push'), 0o755)


@git_group.command()
def push():
    """ Push the current branch and set to track remote. """
    branch = git.current_branch()
    conf.run('git push -u origin {}'.format(branch))


@git_group.command()
def merged(target=None):
    """ Checkout develop, pull and delete merged branches.

    This is to ease the repetitive cleanup of each merged branch.
    """
    protected_branches = ('master', 'develop')
    branch = git.current_branch()

    if target is None:
        target = 'master' if branch.startswith('release/') else 'develop'

    try:
        conf.run('git rev-parse --verify {}'.format(branch))
    except:
        log.err("Branch '{}' does not exist".format(branch))

    log.info("Checking out ^33{}".format(target))
    conf.run('git checkout {}'.format(target))

    log.info("Pulling latest changes")
    conf.run('git pull origin {}'.format(target))

    if branch not in protected_branches:
        log.info("Deleting branch ^33{}".format(branch))
        conf.run('git branch -d {}'.format(branch))

    log.info("Pruning")
    conf.run('git fetch --prune origin')

    log.info("Checking out ^33{}^32 branch".format(target))
    conf.run('git checkout {}'.format(target))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals


from datetime import datetime, timedelta


class CachedPropertyMeta(object):
    def __init__(self, seconds):
        self.last_update = datetime.now() - timedelta(seconds=seconds + 1)
        self.value = None


def cached_property(seconds=1):
    def decorator(fn):
        meta_attr = '__cached_{}'.format(fn.__name__)

        @property
        def prop_wrapper(self):
            meta = self.__dict__.setdefault(meta_attr,
                                            CachedPropertyMeta(seconds))
            since_update = (datetime.now() - meta.last_update).total_seconds()

            if since_update > seconds:
                meta.last_update = datetime.now()
                meta.value = fn(self)

            return meta.value

        return prop_wrapper

    return decorator



# -*- coding: utf-8 -*-
"""
Various commands that do not fit another category but there's not enough of them
to justify a separate module.
"""
from __future__ import absolute_import, unicode_literals, print_function

# stdlib imports
import os
import os.path
import shutil
import sys

# 3rd party imports
import click
try:
    from refdoc import generate_docs as _generate_docs
except ImportError:
    def _generate_docs(*args, **kw):
        print("You need to install sphinx-refdoc to use the 'docs' command",
              file=sys.stderr)
        sys.exit(-1)

# local imports
from peltak.core import log, conf
from . import cli


BUILD_DIR = conf.get_path('BUILD_DIR', '.build')
DOC_SRC_PATH = conf.get_path('DOC_SRC_PATHS', 'docs')
REFDOC_PATHS = conf.get('REFDOC_PATHS', [])

DOC_OUT_PATH = os.path.join(DOC_SRC_PATH, 'html')
DOC_REF_PATH = os.path.join(DOC_SRC_PATH, 'ref')
DOC_ASSETS_PATH = os.path.join(DOC_SRC_PATH, 'assets')
DOC_BUILD_PATH = os.path.join(BUILD_DIR, 'docs')


@cli.command()
@click.option('--recreate', is_flag=True)
@click.option('--no_index', is_flag=False)
def docs(recreate=False, no_index=False):
    """ Build project documentation. """
    log.info('Ensuring assets directory ^94{}^32 exists', DOC_ASSETS_PATH)
    if not os.path.exists(DOC_ASSETS_PATH):
        os.makedirs(DOC_ASSETS_PATH)

    if recreate and os.path.exists(DOC_OUT_PATH):
        log.info("^91Deleting ^94{}".format(DOC_OUT_PATH))
        shutil.rmtree(DOC_OUT_PATH)

    if DOC_REF_PATH:
        _gen_ref_docs(DOC_REF_PATH, no_index)
    else:
        log.err('Not generating any reference documentation - '
                'No DOC_REF_PKG_PATHS specified in config')

    with conf.within_proj_dir(DOC_SRC_PATH):
        log.info('Building docs with ^35sphinx')
        conf.run('sphinx-build -b html -d {build} {docs} {out}'.format(
            build=DOC_BUILD_PATH,
            docs=DOC_SRC_PATH,
            out=DOC_OUT_PATH,
        ))


def _gen_ref_docs(ref_path, no_index=False):
    log.info('Removing previously generated reference documentation')
    if os.path.exists(ref_path):
        shutil.rmtree(ref_path)

    os.makedirs(ref_path)

    log.info('Generating reference documentation')
    args = {'out_dir': ref_path}

    if no_index:
        args['no_index'] = True

    pkg_paths = [conf.proj_path(p) for p in REFDOC_PATHS]

    _generate_docs(pkg_paths, **args)



# -*- coding: utf-8 -*-
"""
TODO:
    - Add optional fields.
        - On creating you specify pairs of name_in_src=field_name, The marker
          must include the field name somehow.
        - On applying during the read, once a marker is encountered the first
          time, the user us prompted for the value. Then that value is used
          for all subsequent occurrences.
"""
from __future__ import absolute_import, unicode_literals

import json
from os.path import abspath, basename
try:
    # python 3
    from io import BytesIO
except ImportError:
    # python 2
    try:
        from cStringIO import StringIO
    except ImportError:
        from StringIO import StringIO

# 3rd party imports
import click
from six import string_types

# local imports
from peltak.core import log
from peltak.core import git
from peltak.core.scaffold import LocalStore, Scaffold
from . import cli


DirPath = click.Path(file_okay=False, dir_okay=True, exists=True)


def marker_def(string):
    name, value = string.split('=')
    name = name.strip()
    value = value.strip()

    if ',' in value:
        value = value.split(',')

    return (name, value)


@cli.group()
def scaffold():
    """ Project scaffold related commands. """
    pass


@scaffold.command()
@click.argument('src_dir', type=DirPath)
@click.option('-n', '--name', type=str)
@click.option('-e', '--exclude', multiple=True, metavar='PATTERN')
@click.option('-m', '--marker', 'markers', type=marker_def, multiple=True, metavar='NAME=VALUE(S)')
@click.option('--no-gitignore', is_flag=True)
def create(src_dir, name, markers, exclude, no_gitignore):
    store = LocalStore()

    markers = dict(markers)
    markers.setdefault('name', basename(abspath(src_dir)))

    if not no_gitignore:
        git_exclude = git.load_gitignore(src_dir)
        exclude = set(exclude) | set(git_exclude) | {'.git'}

    if name is None:
        name = markers['name']
        if not isinstance(name, string_types):
            name = name[0]      # name is a list of names, pick the first one.

    log.info("Creating scaffold ^35{}".format(name))
    log.info("Excluding")
    for pattern in exclude:
        log.info("  ^0- {}", pattern)

    scaffold = Scaffold.create(src_dir, name, exclude, markers)
    store.add(scaffold)


@scaffold.command()
@click.argument('name', type=str, required=True)
@click.option('-f', '--files', 'show_files', is_flag=True)
@click.option('-c', '--config', 'show_config', is_flag=True)
def info(name, show_files, show_config):
    store = LocalStore()
    scaffold = store.load(name)
    # scaffold = Scaffold.load_from_file(name)

    log.info("Name:     ^33{}", scaffold.name)
    log.info("Size:     ^33{} kb", round(scaffold.size / 1024))
    log.info("Created:  ^33{}", scaffold.pretty_created)

    if show_config:
        log.cprint("Config: {")
        for key, value in scaffold.json_config.items():
            log.cprint('  "{name}": ^33{value}^0'.format(
                name=key, value=json.dumps(value, indent=2)
            ))

        log.cprint("}")
        #
        # log.info("Config: ^90{}".format(
        #     json.dumps(scaffold.json_config, indent=2)
        # ))

    if show_files:
        log.info("Files:")
        for path in scaffold.files:
            if path in scaffold.marked_files:
                log.info('  ^0{}', path)
            else:
                log.info('  ^90{}', path)


@scaffold.command()
@click.argument('name', type=str, required=True)
def delete(name):
    log.info("Deleting ^35{}".format(name))
    store = LocalStore()
    store.delete(name)


@scaffold.command()
def list():
    store = LocalStore()

    log.cprint("^32Local:\n^0")
    for scaffold in store.scaffolds:
        log.cprint("  ^90{}  ^0{}", scaffold.pretty_created, scaffold.name)

    log.cprint('')


@scaffold.command()
@click.argument('name', type=str)
@click.argument('proj_name', type=str)
@click.option('-p', '--path', type=DirPath, default='.')
def apply(name, proj_name, path):
    store = LocalStore()
    scaffold = store.load(name)
    scaffold.apply(proj_name, path)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

# stdlib imports
import json
import os
import sys
from datetime import datetime
from os.path import join, isdir, relpath
from zipfile import ZipFile, ZipInfo

try:
    # python 3
    from io import BytesIO
    from zipfile import BadZipFile
except ImportError:
    # python 2
    from zipfile import BadZipfile as BadZipFile
    try:
        from cStringIO import StringIO
    except ImportError:
        from StringIO import StringIO


from six import string_types

# local imports
from peltak.core import log
from peltak.core import fs


TEMPLATE_CONFIG_FILE = 'peltak-template-config.json'

def marker_tag(marker):
    return '_PELTAK-SCAFFOLD-{}_'.format(marker.upper())


class Scaffold(object):
    LINE_SEP = '\n'
    FILE_EXT = '.scaffold'
    CONFIG_FILE = 'peltak-scaffold-config.json'
    TIME_FORMAT = '%Y-%m-%d %H:%M:%S'
    NAME_MARKER = marker_tag('name')

    class Invalid(RuntimeError):
        pass

    def __init__(self, zipfile, **config):
        self.zipfile = zipfile or BytesIO()
        self.config = config
        self.path = None    # Only set if loaded from file or after it's written

        self.config.setdefault('line_sep', Scaffold.LINE_SEP)
        self.config.setdefault('marked_files', [])
        self.config.setdefault('created', datetime.utcnow())
        self.config.setdefault('time_format', Scaffold.TIME_FORMAT)

        if isinstance(self.created, string_types):
            created = datetime.strptime(self.created, self.time_format)
            self.config['created'] = created

    @property
    def name(self):
        return self.config['name']

    @property
    def marked_files(self):
        return self.config['marked_files']

    @property
    def created(self):
        return self.config['created']

    @property
    def pretty_created(self):
        return self.created.strftime(Scaffold.TIME_FORMAT)

    @property
    def time_format(self):
        return self.config['time_format']

    @property
    def size(self):
        start_pos = self.zipfile.tell()

        self.zipfile.seek(0, os.SEEK_END)
        size = self.zipfile.tell()
        self.zipfile.seek(start_pos, os.SEEK_SET)

        return size

    @property
    def json_config(self):
        conf = dict(self.config)
        conf['created'] = self.created.strftime(self.time_format)
        return conf

    @property
    def files(self):
        with ZipFile(self.zipfile) as zip:
            return zip.namelist()

    @classmethod
    def create(cls, src_dir, name, exclude, markers):

        marked_files = []

        zipfile = BytesIO()
        with ZipFile(zipfile, 'w') as zip:
            i = 0
            for path, arc_path in Scaffold._iter_files(src_dir, markers, exclude):
                if isdir(path):
                    # Non-empty dirs will be added through files they contain
                    marked = False
                    # if os.listdir(path) == []:
                    zip.writestr(ZipInfo(arc_path + '/'), '')
                else:
                    content, marked = Scaffold._prepare_file(path, markers)
                    zip.writestr(arc_path, content)

                if marked:
                    marked_files.append(arc_path)

                i += 1
                log.info("[{:4}] Adding ^35{}", i, path)

        return Scaffold(zipfile, name=name, marked_files=marked_files)

    @classmethod
    def load_from_file(cls, path):
        with open(path, 'rb') as fp:
            zipfile = BytesIO(fp.read())

        try:
            with ZipFile(zipfile) as zip:
                data = zip.read(Scaffold.CONFIG_FILE).decode('utf-8')
                config = json.loads(data)
        except BadZipFile as ex:
            raise Scaffold.Invalid(str(ex))

        scaffold = Scaffold(zipfile, **config)
        scaffold.path = path

        return scaffold

    def write(self, path):
        zip_path = join(path, self.name + Scaffold.FILE_EXT)

        with ZipFile(self.zipfile, 'a') as zip:
            zip.writestr(Scaffold.CONFIG_FILE, json.dumps(self.json_config))

        with open(zip_path, 'wb') as fp:
            fp.write(self.zipfile.getvalue())

        self.path = zip_path

    def apply(self, proj_name, out_path):
        with ZipFile(self.zipfile) as zip:
            config = json.loads(zip.read(Scaffold.CONFIG_FILE))

            proj_path = join(out_path, proj_name)
            os.makedirs(proj_path)

            for i, arc_path in enumerate(zip.namelist()):
                if arc_path == Scaffold.CONFIG_FILE:
                    continue

                file_path = join(proj_path, arc_path)

                if Scaffold.NAME_MARKER in arc_path:
                    file_path = file_path.replace(Scaffold.NAME_MARKER,
                                                  proj_name)

                content = zip.read(arc_path)

                if arc_path in config['marked_files']:
                    content = content.decode('utf-8')
                    content = Scaffold._render_file(content, {
                        'name': proj_name
                    })

                    with open(file_path, 'w') as fp:
                        log.cprint("^0[ Template ] ^0^90{}^0", file_path)
                        fp.write(content)
                else:
                    if self._is_dir(zip, arc_path):
                        log.cprint("^34[    Dir   ] ^90{}^0", file_path)
                        os.mkdir(file_path)
                    else:
                        log.cprint("^90[   File   ] ^90{}^0", file_path)
                        with open(file_path, 'wb') as fp:
                            fp.write(content)

    @staticmethod
    def _iter_files(src_dir, markers, exclude):
        for file_name, path in fs.filtered_walk(src_dir, exclude):
            arc_path = relpath(path, src_dir)
            arc_path, _ = Scaffold._prep_line(arc_path, markers)
            # arc_path = arc_path.replace(template_name, Scaffold.NAME_MARKER)

            yield path, arc_path

    @staticmethod
    def _prepare_file(path, markers):
        is_marked = False
        try:
            with open(path) as fp:
                lines = []
                for i, line in enumerate(fp.readlines()):
                    line = line[:-len(os.linesep)]
                    prepped, marked = Scaffold._prep_line(line, markers)

                    if marked:
                        is_marked = True

                    lines.append(prepped)

                return Scaffold.LINE_SEP.join(lines), is_marked

        except UnicodeDecodeError:
            with open(path, 'rb') as fp:
                return fp.read(), False

    @staticmethod
    def _prep_line(line, markers):
        prepped = line
        is_marked = False

        for marker, value in markers.items():
            if isinstance(value, string_types):
                if value in line:
                    prepped = line.replace(value, marker_tag(marker))
                    is_marked = True
            else:
                # Multiple values for marker
                for item in value:
                    if item in line:
                        prepped = line.replace(item, marker_tag(marker))
                        is_marked = True

        return prepped, is_marked

    @staticmethod
    def _render_file(content, values):
        lines = []

        for line in content.split(Scaffold.LINE_SEP):
            rendered = line

            for marker, value in values.items():
                tag = marker_tag(marker)
                if tag in line:
                    rendered = rendered.replace(tag, value)

            lines.append(rendered)

        return os.linesep.join(lines)

    def _is_dir(self, zip, arc_path):
        zip_info = zip.getinfo(arc_path)
        if sys.version_info >= (3, 6):
            return zip_info.is_dir()
        else:
            return zip_info.filename.endswith('/')





import os
import re
from setuptools import setup, find_packages


RE_PY_VERSION = re.compile(
    r'__version__\s*=\s*["\']'
    r'(?P<version>\d+(\.\d+(\.\d+)?)?)'
    r'["\']'
)


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def read_version():
    content = read('src/peltak/__init__.py')
    m = RE_PY_VERSION.search(content)
    if not m:
        return '0.0'
    else:
        return m.group('version')


setup(
    name="peltak",
    version=read_version(),
    author="Mateusz 'novo' Klos",
    author_email="novopl@gmail.com",
    license="MIT",
    keywords="fabric ops devops",
    url="http://github.com/novopl/peltak",
    description="Set of CLI commands to help manage a project",
    long_description=read('README.rst'),
    package_dir={'': 'src'},
    packages=find_packages('src'),
    install_requires=[
        l.strip() for l in read('requirements.txt').split() if '==' in l
    ],
    entry_points={
        'console_scripts': [
            'peltak = peltak.main:cli',
        ]
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "License :: OSI Approved :: MIT License",
    ],
)
